{%- extends "sphinx_rtd_theme/breadcrumbs.html" %}

{# We use framagit and not gitlab, and we don't want to add a link on generated pages #}

{% block breadcrumbs_aside %}
  <li class="wy-breadcrumbs-aside">
    {% if hasdoc(pagename) and pagename != "search" and pagename != "genindex" %}
    {# TODO: How to rule out all java pages?
     #    (not re.match(pagename, "java/org/simgrid"))
     # fails because re is not imported #}

      <a href="{{ meta['framagit_url'] }}" class="fa fa-bitbucket"> {{ _('Edit on FramaGit') }}</a>
    {% endif %}
  </li>
{% endblock %}
