/* Copyright (c) 2021-2022. The SimGrid Team. All rights reserved.          */

/* This program is free software; you can redistribute it and/or modify it
 * under the terms of the license (GNU LGPL) which comes with this package. */

#include <simgrid/plugins/ProducerConsumer.hpp>

XBT_LOG_NEW_CATEGORY(producer_consumer, "Producer-Consumer plugin logging category");

namespace simgrid {
namespace plugin {
unsigned long pc_id = 0;
}
}
