/* module - modularize the code                                             */

/* Copyright (c) 2004-2022. The SimGrid Team. All rights reserved.          */

/* This program is free software; you can redistribute it and/or modify it
 * under the terms of the license (GNU LGPL) which comes with this package. */

#ifndef XBT_MODULE_H
#define XBT_MODULE_H

#include <xbt/misc.h>           /* XBT_PUBLIC */

SG_BEGIN_DECL

XBT_PUBLIC void xbt_init(int* argc, char** argv);

SG_END_DECL

#endif /* XBT_MODULE_H */
